﻿/**
 * copyright (c) 2010 IxD Consultant (http://www.ixdc.net)
 * license
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License <http://www.gnu.org/copyleft/gpl.html> for more details.
   
 * version: 1.1
 * date: 19/03/2016
 * Actionscript 2.0
 * Flash Player 8
 * @author Ahmed Abbas, a.abbas@ixdc.net
 *
 * requirements:
 
 	1. dynamic TextField.
	2. HTML enabled TextField.
	3. pre-assigned TextFormat.
	4. Arabic fonts.
	
 * 
 * properties:
 
 	1. data: for referencing original input string.
	2. wrapFactor: optional reduction value (Mac only) for wrapping correction when using bi-directional text  (default = .99).
	4. embedFonts: boolean (default = true).
	5. hindiNumeralsOnly: boolean (default = false).
 
 *
 * features supported:
 
 	1.  embedding fonts (just put a dynamic textfield on-stage and select at least Basic Latin (95 glyphs) and Arabic (1088 glyphs) from the Character Embedding menu).
	2.  Arabic ligatures.
	3.  Arabic diacritics.
	4.  word wrapping.
	5.  bi-directional text.
	7.  loading external text on run-time.
	8.  Windows/Mac/Linux support.
	9.  Arabic enabled input fields using actionscript method (createArabicInput) with the help of javascript (arabicinput.js).
	10. Urdu support.
 
 *
 * example:
 	
	import net.ixdc.utils.ArabicLight;
	var utils:ArabicLight = new ArabicLight();
	utils.wrapFactor = 1.1; // optional
	
	var format:TextFormat = new TextFormat();
	format.font = "Arial";
	format.size = 14;
	format.align = "right";
	format.rightMargin = 4;
	
	var output:TextField = this.createTextField("output", 1, 10, 10, Stage.width-20, format.size);
	output.autoSize = true;
	output.embedFonts = true;
	output.wordWrap = true;
	output.multiline = true;
	output.html = true;
	
	var xml:XML = new XML();
	xml.ignoreWhite = true;
	xml.onLoad = function(done:Boolean) {
		if (done) {
			output.htmlText = utils.parseArabic(this.firstChild.firstChild.nodeValue, output._width, format);
		}
	}
	xml.load("arabic.xml");
	
	<?xml version="1.0" encoding="utf-8"?>
	<arabic>
		<![CDATA[الهدف من لعبة الكلمات المتقاطعة هو ملء المربعات البيضاء، وتشكيل الكلمات أو العبارات، عن طريق حل القرائن التي تؤدي إلى إجابات. والمربعات السوداء تستخدم لفصل الكلمات أو العبارات. ويوضع لكل صف أو عمود رقم، ثم يكتب أمام الرقم ما يشابه الكلمة المطلوب كتابتها في المربعات، وتحوي المجلات أنواعا من هذه اللعبة؛ فمنها سهل وآخر صعب، والكلمات المتقاطعة في مجلة معينة تتبع أسلوب يختلف عن الأخرى، تبعا لأسلوب اللغة المستعملة والبلد، وتقوم المجلات والصحف بوضع هذه اللعبة لتسلية القراء. ظهرت أول لعبة للكلمات المتقاطعة في صحيفة newyork world وذلك في 21 ديسمبر عام 1913، وأصبحت من الألعاب الرائجة في الولايات المتحدة، ومنها انتقلت إلى بقية دول العالم، وبمختلف اللغات، وكان أول من أدخلها إلى الصحافة هو آرثر وين.]]>
	</arabic>
 
 **/
import flash.external.ExternalInterface;
import net.ixdc.utils.RegExp;
class net.ixdc.utils.ArabicLight {
	
	public var data:String;
	public var wrapFactor:Number = .99; // optional reduction for wrapping correction when using bi-directional text
	public var embedFonts:Boolean = true;
	public var hindiNumeralsOnly:Boolean = false;
	
	private var tag_rgx:RegExp;
	private var specialChars:String = "";
	private var arabicPunctuation:String;
	private var isMAC:Boolean;
	private var chars:String;
	
	function ArabicLight() {
		// check for OS, it seems like Flash Player behaves differently between Windows and Mac, especially for embedding fonts!
		isMAC = false;
		var version:String = getVersion();
		if (version.toLowerCase().indexOf("mac") != -1) {
			isMAC = true;
		}
		tag_rgx = new RegExp('<(?\\s|[^>]*)>', 'gim');
		specialChars += String.fromCharCode(1575, 1571, 1573, 1570, 1583, 1672, 1584, 1585, 1681, 1586, 1688, 1608, 1572, 1569);
		arabicPunctuation = String.fromCharCode(0x0609, 0x060A, 0x061B, 0x061E, 0x061F, 0x060C, 0x060D, 0x066A, 0x066B, 0x066C, 0x066D, 0x06D4);
	}
	
	/**
	 * @public
	 * creates proper arabic HTML string.
	 * 
	 * @param input String required value of original arabic string.
	 * @param lineWidth Number required value of wrapping line width.
	 * @param fieldFormat TextFormat optional value of final text format.
	 * @return String value with correct arabic text.
	 */
	public function parseArabic(input:String, lineWidth:Number, format:TextFormat):String {
		
		if (!format) {
			format = new TextFormat();
			format.font = 'Arial';
			format.size = 12;
			format.align = 'right';
			format.rightMargin = 4;
		}
		
		// referencing original input string to use for matching/comparing instead of (TextField.text)
		data = input;
		
		var string:String = input;
		var i:Number;
		var v:Number;
		
		var time:Date = new Date();
		var begin:Number = time.getTime();
		var end:Number;
		
		// replace html line breaks
		string = removeLineBreaks(string);
		
		// explode bulleted list
		string = splitBulletList(string);
		
		// remove html tags
		var str = string;
		string = str.replace(tag_rgx, '');
		//trace(string);
		//trace(string.length+' characters');
		time = new Date();
		end = time.getTime();
		trace('ArabicLight > parseArabic | AS2 v1.1\n');
		trace('embedFonts: '+embedFonts+'\n');
		trace('stripped html tags in '+(end-begin)+' ms\n');
		
		var numChars:Number = string.length;
		var words:Array;
		var numWords:Number;
		var word:String;
		var tempLine:String = '';
		var breakLine:Boolean = false;
		
		// split words for manual wrapping
		words = string.split(' ');
		numWords = words.length;
		
		// temporarily create an autosized textfield off-stage to use for proper right to left wrapping
		// if using V2 Components somewhere, then use a static depth instead of _root.getNextHighestDepth()
		var temp:TextField = _root.createTextField("temp", _root.getNextHighestDepth(), -1000, -1000, 10, format.size+4);
		temp.autoSize = true;
		temp.embedFonts = embedFonts;
		temp.selectable = false;
		temp.html = true;
		
		string = '';
		i = 0;
		while (i<numWords) {
			word = words[i];
			if (string != '') {
				string += ' ';
			}
			string += word;
			// resize temporarily textfield
			if (tempLine != '') {
				tempLine += ' ';
			}
			tempLine += word;
			/*if (validateArabicWord(word)) {
				// replace arabic characters from unicode charts at the temporarily textfield for proper line width measurement
				tempLine += replaceArabic(word);
			} else {
				tempLine += word;
			}*/
			temp.text = tempLine;
			temp.setTextFormat(format);
			// measure max width for proper text wrapping
			breakLine = false;
			if (isMAC) {
				var metrics:Object = format.getTextExtent(tempLine);
				if (metrics.width>=(lineWidth*wrapFactor)-(format.leftMargin ? Number(format.leftMargin) : 0)-(format.rightMargin ? Number(format.rightMargin) : 0)) {
					breakLine = true;
				}
			} else {
				if (Math.ceil(temp._width)>=(lineWidth*wrapFactor)-(format.leftMargin ? Number(format.leftMargin) : 0)-(format.rightMargin ? Number(format.rightMargin) : 0)) {
					breakLine = true;
				}
			}
			// break text line
			if (breakLine) {
				string += '\n';
				tempLine = '';
			}
			i++;
		}
		
		// process wrapped text lines
		string = processTextLines(string);
		
		// remove temporarily textfield
		temp.removeTextField();
		
		// replace chevrons
		string = string.split('<').join('&lt;');
		string = string.split('>').join('&gt;');
		
		// flip brackets
		string = flipBrackets(string);
		
		// fix some bugs !
		string = fixBugs(string);
		
		// convert numbers to hindi
		if (hindiNumeralsOnly) {
			string = convertNumbers(string);
		}
		
		// convert text format to html tags
		string = getHTMLFormat(string, format);
		
		time = new Date();
		end = time.getTime();
		trace('arabic characters has been replaced in: '+(end-begin)+' ms\n');
		
		//trace(string);
		time = new Date();
		end = time.getTime();
		trace('parse complete in '+(end-begin)+' ms');
		trace('-------------------------------------------------\n');
		
		// returning proper arabic text
		return string;
	}
	
	/**
	 * @public
	 * creates proper arabic input field.
	 * 
	 * @param flashID String swf object id in HTML document.
	 * @param inputID String generated HTML input id.
	 * @param field TextField target text input in flash.
	 * @param color Number optional value for HTML input color.
	 */
	public function createArabicInput(flashID:String, inputID:String, field:TextField, color:Number, bgColor:Number):Void {
		var c:String = "#000000";
		if (color != undefined) {
			c = getProperColor(color);
		}
		var b:String = "none";
		if (bgColor != undefined) {
			b = getProperColor(bgColor);
		}
		var format:TextFormat = field.getTextFormat();
		var point:Object = {x:field._x, y:field._y};
		field._parent.localToGlobal(point);
		var type:String = "singleline";
		if (field.multiline) {
			type = "multiline";
		}
		ExternalInterface.call("createArabicInput", flashID, inputID, field.text, point.x, point.y, field._width, field._height, c, b, type);
	}
	
	/**
	 * @public
	 * reveal all generated arabic input fields.
	 */
	public function showArabicInput():Void {
		ExternalInterface.call("revealAllArabicInputs");
	}
	
	private function processTextLines(input:String):String {
		var lines:Array = input.split('\n');
		var numLines:Number = lines.length;
		var numLineChars:Number;
		var line:String;
		var code:Number;
		var validArabicChar:Boolean;
		var validSymbol:Boolean;
		var toggleArabic:Boolean;
		var arabicString:String = '';
		var noneArabicString:String = '';
		var tempChunk:Array;
		var lineChunks:Array;
		
		var string:String = '';
		var i:Number = 0;
		var v:Number;
		
		while (i<numLines) {
			lineChunks = [];
			line = lines[i];
			numLineChars = line.length;
			v = 0;
			while (v<numLineChars) {
				code = line.charCodeAt(v);
				validArabicChar = (code >= 1536 && code <= 1631) || 
									(code >= 1642 && code <= 1791) || 
									(code >= 65136 && code <= 65279) || 
									validateSymbol(line, v);
				if (validArabicChar) {
					toggleArabic = true;
				} else {
					toggleArabic = false;
				}
				if (toggleArabic) {
					if (validateSymbol(line, v) && !validateNextArabicChar(line, v) && noneArabicString != "") {
						toggleArabic = false;
					}
				}
				if (toggleArabic) {
					if (noneArabicString != '') {
						lineChunks.push(noneArabicString);
						noneArabicString = '';
					}
					arabicString += line.charAt(v);
				} else {
					if (arabicString != '') {
						// replace arabic characters from unicode charts
						arabicString = replaceArabic(arabicString);
						if (embedFonts) {
							// reverse arabic characters (for embedFonts only)
							tempChunk = arabicString.split('');
							tempChunk.reverse();
							arabicString = tempChunk.join('');
						}
						lineChunks.push(arabicString);
						arabicString = '';
					}
					noneArabicString += line.charAt(v);
				}
				v++;
			}
			// collect remaining string chunks !
			if (arabicString != '') {
				// replace arabic characters from unicode charts
				arabicString = replaceArabic(arabicString);
				if (embedFonts) {
					// reverse arabic characters (for embedFonts only)
					tempChunk = arabicString.split('');
					tempChunk.reverse();
					arabicString = tempChunk.join('');
				}
				lineChunks.push(arabicString);
				arabicString = '';
			}
			if (noneArabicString != '') {
				lineChunks.push(noneArabicString);
				noneArabicString = '';
			}
			if (embedFonts) {
				// reverse line chunks (for embedFonts only)
				lineChunks.reverse();
			}
			string += lineChunks.join('');
			if (i<numLines-1) {
				string += '\n';
			}
			i++;
		}
		return string;
	}
	
	private function replaceArabic(input:String):String {
	
		chars = input;
		
		var string:String = '';
		var numChars:Number = chars.length;
		var i:Number = 0;
		
		while (i<numChars) {
			string += replaceArabicChar(i);
			i++;
		}
		
		return string;
		
	}
	
	private function replaceArabicChar(i:Number):String {
		
		var string:String;
		switch (chars.charCodeAt(i)) {
			case 1575: //    ا
			string = setChar(i, String.fromCharCode(0x0627), String.fromCharCode(0x0627), String.fromCharCode(0xFE8E), String.fromCharCode(0xFE8E));
			break;
			case 1571: //    أ
			string = setChar(i, String.fromCharCode(0x0623), String.fromCharCode(0x0623), String.fromCharCode(0xFE84), String.fromCharCode(0xFE84));
			break;
			case 1573: //    إ
			string = setChar(i, String.fromCharCode(0x0625), String.fromCharCode(0x0625), String.fromCharCode(0xFE88), String.fromCharCode(0xFE88));
			break;
			case 1570: //    آ
			string = setChar(i, String.fromCharCode(0x0622), String.fromCharCode(0x0622), String.fromCharCode(0xFE82), String.fromCharCode(0xFE82));
			break;
			case 1576: //    ب
			string = setChar(i, String.fromCharCode(0x0628), String.fromCharCode(0xFE91), String.fromCharCode(0xFE92), String.fromCharCode(0xFE90));
			break;
			case 1662: //    پ
			string = setChar(i, String.fromCharCode(0x067E), String.fromCharCode(0xFB58), String.fromCharCode(0xFB59), String.fromCharCode(0xFB57));
			break;
			case 1578: //    ت
			string = setChar(i, String.fromCharCode(0x062A), String.fromCharCode(0xFE97), String.fromCharCode(0xFE98), String.fromCharCode(0xFE96));
			break;
			case 1657: //    ٹ
			string = setChar(i, String.fromCharCode(0x0679), String.fromCharCode(0xFB68), String.fromCharCode(0xFB69), String.fromCharCode(0xFB67));
			break;
			case 1579: //    ث
			string = setChar(i, String.fromCharCode(0x062B), String.fromCharCode(0xFE9B), String.fromCharCode(0xFE9C), String.fromCharCode(0xFE9A));
			break;
			case 1580: //    ج
			string = setChar(i, String.fromCharCode(0x062C), String.fromCharCode(0xFE9F), String.fromCharCode(0xFEA0), String.fromCharCode(0xFE9E));
			break;
			case 1670: //    چ
			string = setChar(i, String.fromCharCode(0x686), String.fromCharCode(0xFB7C), String.fromCharCode(0xFB7D), String.fromCharCode(0xFB7B));
			break;
			case 1581: //    ح
			string = setChar(i, String.fromCharCode(0x062D), String.fromCharCode(0xFEA3), String.fromCharCode(0xFEA4), String.fromCharCode(0xFEA2));
			break;
			case 1582: //    خ
			string = setChar(i, String.fromCharCode(0x062E), String.fromCharCode(0xFEA7), String.fromCharCode(0xFEA8), String.fromCharCode(0xFEA6));
			break;
			case 1583: //    د
			string = setChar(i, String.fromCharCode(0x062F), String.fromCharCode(0x062F), String.fromCharCode(0xFEAA), String.fromCharCode(0xFEAA));
			break;
			case 1672: //    ڈ
			string = setChar(i, String.fromCharCode(0x0688), String.fromCharCode(0xFB88), String.fromCharCode(0xFB89), String.fromCharCode(0xFB89));
			break;
			case 1584: //    ذ
			string = setChar(i, String.fromCharCode(0x0630), String.fromCharCode(0x0630), String.fromCharCode(0xFEAC), String.fromCharCode(0xFEAC));
			break;
			case 1585: //    ر
			string = setChar(i, String.fromCharCode(0x0631), String.fromCharCode(0x0631), String.fromCharCode(0xFEAE), String.fromCharCode(0xFEAE));
			break;
			case 1681: //    ڑ
			string = setChar(i, String.fromCharCode(0x0691), String.fromCharCode(0xFB8C), String.fromCharCode(0xFB8D), String.fromCharCode(0xFB8D));
			break;
			case 1586: //    ز
			string = setChar(i, String.fromCharCode(0x0632), String.fromCharCode(0x0632), String.fromCharCode(0xFEB0), String.fromCharCode(0xFEB0));
			break;
			case 1688: //    ژ
			string = setChar(i, String.fromCharCode(0x0698), String.fromCharCode(0xFB8A), String.fromCharCode(0xFB8B), String.fromCharCode(0xFB8B));
			break;
			case 1587: //    س
			string = setChar(i, String.fromCharCode(0x0633), String.fromCharCode(0xFEB3), String.fromCharCode(0xFEB4), String.fromCharCode(0xFEB2));
			break;
			case 1588: //    ش
			string = setChar(i, String.fromCharCode(0x0634), String.fromCharCode(0xFEB7), String.fromCharCode(0xFEB8), String.fromCharCode(0xFEB6));
			break;
			case 1589: //    ص
			string = setChar(i, String.fromCharCode(0x0635), String.fromCharCode(0xFEBB), String.fromCharCode(0xFEBC), String.fromCharCode(0xFEBA));
			break;
			case 1590: //    ض
			string = setChar(i, String.fromCharCode(0x0636), String.fromCharCode(0xFEBF), String.fromCharCode(0xFEC0), String.fromCharCode(0xFEBE));
			break;
			case 1591: //    ط
			string = setChar(i, String.fromCharCode(0x0637), String.fromCharCode(0xFEC3), String.fromCharCode(0xFEC4), String.fromCharCode(0xFEC2));
			break;
			case 1592: //    ظ
			string = setChar(i, String.fromCharCode(0x0638), String.fromCharCode(0xFEC7), String.fromCharCode(0xFEC8), String.fromCharCode(0xFEC6));
			break;
			case 1593: //    ع
			string = setChar(i, String.fromCharCode(0x0639), String.fromCharCode(0xFECB), String.fromCharCode(0xFECC), String.fromCharCode(0xFECA));
			break;
			case 1594: //    غ
			string = setChar(i, String.fromCharCode(0x063A), String.fromCharCode(0xFECF), String.fromCharCode(0xFED0), String.fromCharCode(0xFECE));
			break;
			case 1601: //    ف
			string = setChar(i, String.fromCharCode(0x0641), String.fromCharCode(0xFED3), String.fromCharCode(0xFED4), String.fromCharCode(0xFED2));
			break;
			case 1700: //    ڤ
			string = setChar(i, String.fromCharCode(0xFB6A), String.fromCharCode(0xFB6C), String.fromCharCode(0xFB6D), String.fromCharCode(0xFB6B));
			break;
			case 1602: //    ق
			string = setChar(i, String.fromCharCode(0x0642), String.fromCharCode(0xFED7), String.fromCharCode(0xFED8), String.fromCharCode(0xFED6));
			break;
			case 1603: //    ك
			string = setChar(i, String.fromCharCode(0x0643), String.fromCharCode(0xFEDB), String.fromCharCode(0xFEDC), String.fromCharCode(0xFEDA));
			break;
			case 1705: //    ک
			string = setChar(i, String.fromCharCode(0x06A9), String.fromCharCode(0xFB90), String.fromCharCode(0xFB91), String.fromCharCode(0xFB8F));
			break;
			case 1711: //    گ
			string = setChar(i, String.fromCharCode(0x06AF), String.fromCharCode(0xFB94), String.fromCharCode(0xFB95), String.fromCharCode(0xFB93));
			break;
			case 1604: //    ل
			string = setChar(i, String.fromCharCode(0x0644), String.fromCharCode(0xFEDF), String.fromCharCode(0xFEE0), String.fromCharCode(0xFEDE));
			break;
			case 1605: //    م
			string = setChar(i, String.fromCharCode(0x0645), String.fromCharCode(0xFEE3), String.fromCharCode(0xFEE4), String.fromCharCode(0xFEE2));
			break;
			case 1606: //    ن
			string = setChar(i, String.fromCharCode(0x0646), String.fromCharCode(0xFEE7), String.fromCharCode(0xFEE8), String.fromCharCode(0xFEE6));
			break;
			case 1722: //    ں
			string = setChar(i, String.fromCharCode(0x06BA), String.fromCharCode(0xFB9E), String.fromCharCode(0xFB9F), String.fromCharCode(0xFB9F));
			break;
			case 1723: //    ڻ
			string = setChar(i, String.fromCharCode(0x06BB), String.fromCharCode(0xFBA2), String.fromCharCode(0xFBA3), String.fromCharCode(0xFBA1));
			break;
			case 1607: //    ه
			string = setChar(i, String.fromCharCode(0x0647), String.fromCharCode(0xFEEB), String.fromCharCode(0xFEEC), String.fromCharCode(0xFEEA));
			break;
			case 1726: //    ھ
			string = setChar(i, String.fromCharCode(0xFBAA), String.fromCharCode(0xFBAC), String.fromCharCode(0xFBAD), String.fromCharCode(0xFBAB));
			break;
			case 1729: //    ہ
			string = setChar(i, String.fromCharCode(0xFBA6), String.fromCharCode(0xFBA8), String.fromCharCode(0xFBA9), String.fromCharCode(0xFBA7));
			break;
			case 1577: //    ة
			string = setChar(i, String.fromCharCode(0x0629), "", "", String.fromCharCode(0xFE94));
			break;
			case 1608: //    و
			string = setChar(i, String.fromCharCode(0x0648), String.fromCharCode(0x0648), String.fromCharCode(0xFEEE), String.fromCharCode(0xFEEE));
			break;
			case 1572: //    ؤ
			string = setChar(i, String.fromCharCode(0x0624), String.fromCharCode(0x0624), String.fromCharCode(0xFE86), String.fromCharCode(0xFE86));
			break;
			case 1609: //    ى
			string = setChar(i, String.fromCharCode(0x0649), String.fromCharCode(0x0649), String.fromCharCode(0xFEF0), String.fromCharCode(0xFEF0));
			break;
			case 1746: //    ے
			string = setChar(i, String.fromCharCode(0x06D2), String.fromCharCode(0xFBAE), String.fromCharCode(0xFBAF), String.fromCharCode(0xFBAF));
			break;
			case 1610: //    ي
			case 1740: // TODO: caught during urdu support!
			string = setChar(i, String.fromCharCode(0x064A), String.fromCharCode(0xFEF3), String.fromCharCode(0xFEF4), String.fromCharCode(0xFEF2));
			break;
			case 1574: //     ئ
			string = setChar(i, String.fromCharCode(0x0626), String.fromCharCode(0xFE8B), String.fromCharCode(0xFE8C), String.fromCharCode(0xFE8A));
			break;
			case 1747: //    ۓ
			string = setChar(i, String.fromCharCode(0x06D3), String.fromCharCode(0xFBB0), String.fromCharCode(0xFBB1), String.fromCharCode(0xFBB1));
			break;
			case 1569: //    ء
			string = String.fromCharCode(0x0621);
			break;
			case 1600: //    ـ
			string = String.fromCharCode(0x0640);
			break;
			case 1614: //    َ
			string = setChar(i, String.fromCharCode(0xFE76), String.fromCharCode(0xFE76), String.fromCharCode(0xFE77), String.fromCharCode(0xFE76));
			break;
			case 1611: //    ً
			string = setChar(i, String.fromCharCode(0xFE70), String.fromCharCode(0xFE70), String.fromCharCode(0xFE71), String.fromCharCode(0xFE70));
			break;
			case 1616: //    ِ
			string = setChar(i, String.fromCharCode(0xFE7A), String.fromCharCode(0xFE7A), String.fromCharCode(0xFE7B), String.fromCharCode(0xFE7A));
			break;
			case 1613: //    ٍ
			string = String.fromCharCode(0xFE74);
			break;
			case 1615: //    ُ
			string = setChar(i, String.fromCharCode(0xFE78), String.fromCharCode(0xFE78), String.fromCharCode(0xFE79), String.fromCharCode(0xFE78));
			break;
			case 1612: //    ٌ
			string = String.fromCharCode(0xFE72);
			break;
			case 1618: //    ْ
			string = setChar(i, String.fromCharCode(0xFE7E), String.fromCharCode(0xFE7E), String.fromCharCode(0xFE7F), String.fromCharCode(0xFE7E));
			break;
			case 1617: //    ّ
			string = setChar(i, String.fromCharCode(0xFE7C), String.fromCharCode(0xFE7C), String.fromCharCode(0xFE7F), String.fromCharCode(0xFE7C));
			break;
			case 63: //    ?
			string = String.fromCharCode(0x061F);
			break;
			case 44: //    ,
			string = String.fromCharCode(0x060C);
			break;
			case 59: //    ;
			string = String.fromCharCode(0x061B);
			break;
			case 37: //    %
			string = String.fromCharCode(0x066A);
			break;
			default:
			string = chars.charAt(i);
			break;
		}
		
		return string;
		
	}
	
	private function setChar(i:Number, solo:String, begin:String, middle:String, end:String):String {
		
		var string:String = "";
		
		// detect lam-alef (ﻻ) cases
		if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "ا") {
			if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
				string = String.fromCharCode(0xFEFC);
			} else {
				string = String.fromCharCode(0xFEFB);
			}
			chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
		} else if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "أ") {
			if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
				string = String.fromCharCode(0xFEF8);
			} else {
				string = String.fromCharCode(0xFEF7);
			}
			chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
		} else if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "إ") {
			if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
				string = String.fromCharCode(0xFEFA);
			} else {
				string = String.fromCharCode(0xFEF9);
			}
			chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
		} else if (chars.charAt(i) == "ل" && chars.charAt(i+1) == "آ") {
			if (validateArabicChar(i-1) && specialChars.indexOf(chars.charAt(i-1)) == -1) {
				string = String.fromCharCode(0xFEF6);
			} else {
				string = String.fromCharCode(0xFEF5);
			}
			chars = chars.substring(0, i)+string+chars.substring(i+2, chars.length);
		} else {
			// check for arabic character position in word (solo, begin, middle, end)
			var isDiacriticPrev:Boolean = chars.charCodeAt(i-1) >= 1611 && chars.charCodeAt(i-1) <= 1618;
			var isDiacriticNext:Boolean = chars.charCodeAt(i+1) >= 1611 && chars.charCodeAt(i+1) <= 1618;
			if (i == 0) {
				if (specialChars.indexOf(chars.charAt(i)) != -1 || !validateArabicChar(i+1)) {
					string = solo;
				} else {
					string = begin;
				}
			} else if (i == chars.length-1) {
				if (specialChars.indexOf(chars.charAt(i-1)) != -1 || !validateArabicChar(i-1)) {
					string = solo;
				} else {
					string = end;
				}
			} else if (validateArabicChar(i-1) && validateArabicChar(i+1)) {
				if (specialChars.indexOf(chars.charAt(i-1)) != -1) {
					if (specialChars.indexOf(chars.charAt(i)) != -1 || isDiacriticNext && !validateArabicChar(i+2)) {
						string = solo;
					} else {
						string = begin;
					}
				} else {
					if (specialChars.indexOf(chars.charAt(i)) != -1 || chars.charAt(i+1) == "ء" || chars.charAt(i) == "ة") {
						if (chars.charAt(i-1) != "ة") {
							if (isDiacriticPrev && specialChars.indexOf(chars.charAt(i-2)) != -1) {
								string = begin;
							} else {
								string = end;
							}
						} else {
							string = begin;
						}
					} else {
						if (chars.charAt(i-1) != "ة") {
							if (isDiacriticPrev && specialChars.indexOf(chars.charAt(i-2)) != -1) {
								string = begin;
							} else if (isDiacriticNext && !validateArabicChar(i+2)) {
								string = end;
							} else {
								if (arabicPunctuation.indexOf(chars.charAt(i+1)) != -1) {
									string = end;
								} else {
									string = middle;
								}
							}
						} else {
							string = begin;
						}
					}
				}
			} else {
				if (validateArabicChar(i-1) && !validateArabicChar(i+1)) {
					if (specialChars.indexOf(chars.charAt(i-1)) != -1 || isDiacriticPrev && specialChars.indexOf(chars.charAt(i-2)) != -1) {
						string = solo;
					} else {
						string = end;
					}
				} else if (!validateArabicChar(i-1) && validateArabicChar(i+1)) {
					if (specialChars.indexOf(chars.charAt(i)) != -1) {
						string = solo;
					} else {
						string = begin;
					}
				} else if (!validateArabicChar(i-1) && !validateArabicChar(i+1)) {
					string = solo;
				}
			}
		}
		
		return string;
		
	}
	
	private function validateArabicChar(i:Number):Boolean {
		
		var valid:Boolean = false;
		var code:Number = chars.charCodeAt(i);
		// Arabic Default
		var arabicDefaultRange:Boolean = code >= String.fromCharCode(0x0600).charCodeAt(0) && code <= String.fromCharCode(0x06FF).charCodeAt(0);
		// Arabic Presentation Forms-A
		var arabicPresentationFormsARange:Boolean = code >= String.fromCharCode(0xFB50).charCodeAt(0) && code <= String.fromCharCode(0xFDFF).charCodeAt(0);
		// Arabic Presentation Forms-B
		var arabicPresentationFormsBRange:Boolean = code >= String.fromCharCode(0xFE70).charCodeAt(0) && code <= String.fromCharCode(0xFEFF).charCodeAt(0);
		// Arabic Extended-A
		var arabicExtendedARange:Boolean = code >= String.fromCharCode(0x08A0).charCodeAt(0) && code <= String.fromCharCode(0x08FF).charCodeAt(0);
		// Arabic Supplement
		var arabicSupplementRange:Boolean = code >= String.fromCharCode(0x0750).charCodeAt(0) && code <= String.fromCharCode(0x077F).charCodeAt(0);
		if (i>=0 && i<chars.length) {
			if (arabicDefaultRange || arabicPresentationFormsARange || arabicPresentationFormsBRange || arabicExtendedARange || arabicSupplementRange) {
				valid = true;
			}
		}
		return valid;
		
	}
	
	private function validateArabicWord(input:String, i:Number):Boolean {
		
		if (!i) {
			i = 0;
		}
		
		var max:Number = input.length-2;
		var code:Number = input.charCodeAt(i);
		var valid:Boolean = (code >= 1536 && code <= 1791 || code >= 65136 && code <= 65279);
		
		if (!valid && i<max) {
			valid = validateArabicWord(input, i+1);
		}
		
		return valid;
		
	}
	
	private function validateNextArabicChar(input:String, i:Number):Boolean {
		
		var code:Number = input.charCodeAt(i);
		var valid:Boolean = (code >= 1536 && code <= 1791 || code >= 65136 && code <= 65279);
		
		if (!valid) {
			if (input.charAt(i) == ' ' || validateSymbol(input, i)) {
				valid = validateNextArabicChar(input, i+1);
			}
		}
		
		return valid;
		
	}
	
	private function validateSymbol(input:String, i:Number):Boolean {
		
		var valid:Boolean = false;
		var code:Number = input.charCodeAt(i);
		
		if (code >= 32 && code <= 47 || code >= 58 && code <= 63 || code >= 91 && code <= 93 || code >= 123 && code <= 126 || code == 163 || code == 171 || code == 187 || code == 1548 || code == 8226 || code == 9642 || code == 10216 || code == 10217) {
			valid = true;
		}
		
		return valid;
		
	}
	
	private function cleanString(input:String):String {
		var string:String = "";
		var numInput:Number = input.length;
		var i:Number = 0;
		while (i<numInput) {
			// ASCII control characters charcode range 0-31 and charcode 127
			// Reference: http://en.wikipedia.org/wiki/ASCII#ASCII_control_characters
			if (input.charCodeAt(i)>31 && input.charCodeAt(i) != 127) {
				string += input.charAt(i);
			}
			i++;
		}
		return string;
	}
	
	private function removeLineBreaks(input:String):String {
		input = input.split('<br />').join('\n');
		input = input.split('<br>').join('\n');
		input = input.split('<BR />').join('\n');
		input = input.split('<BR>').join('\n');
		return input;
	}
	
	private function splitBulletList(input:String):String {
		input = input.split('<ul>').join('');
		input = input.split('</ul>').join('\n');
		input = input.split('<li>').join('\n• ');
		input = input.split('</li>').join('');
		return input;
	}
	
	private function flipBrackets(input:String):String {
		
		// replace brackets with codes
		input = input.split('[').join('#LSB#');
		input = input.split(']').join('#RSB#');
		input = input.split('{').join('#LCB#');
		input = input.split('}').join('#RCB#');
		input = input.split('(').join('#LPT#');
		input = input.split(')').join('#RPT#');
		
		// replace code with flipped brackets
		input = input.split('#LSB#').join(']');
		input = input.split('#RSB#').join('[');
		input = input.split('#LCB#').join('}');
		input = input.split('#RCB#').join('{');
		input = input.split('#LPT#').join(')');
		input = input.split('#RPT#').join('(');
		
		return input;
		
	}
	
	private function fixBugs(input:String):String {
		
		var string = '';
		var lines:Array = input.split('\n');
		var numLines:Number = lines.length;
		var line:String;
		var numLineChars:Number;
		var tempLine:Array;
		var code:Number;
		var validNextChar:Boolean;
		var singleBracket:Object;
		var validArabicChar:Boolean;
		var numbericString:String = '';
		var numericSeparator:String = ', ';
		var i:Number = 0;
		var v:Number;
		
		while (i<numLines) {
			
			line = lines[i];
			numLineChars = line.length;
			
			if (!embedFonts) {
				
				// fix bullet position with bi-directional line (for none-embedFonts only) !
				if (line.charAt(0) == '•') {
					code = line.charCodeAt(2);
					validNextChar = (code >= 1536 && code <= 1791 || code >= 65136 && code <= 65279);
					if (!validNextChar) {
						line = line.slice(1, line.length)+' •';
					}
				}
				
				// fix single bracket occurance with bi-directional line (for none-embedFonts only) !
				singleBracket = getSingleBracket(line);
				if (singleBracket.bracket) {
					tempLine = line.split(singleBracket.bracket);
					validArabicChar = (line.charCodeAt(0) >= 1536 && line.charCodeAt(0) <= 1791 || 
									   line.charCodeAt(0) >= 65136 && line.charCodeAt(0) <= 65279);
					if (!validArabicChar && !isNumeric(line.charAt(0)) && !singleBracket.closed) {
						tempLine.splice(1, 0, singleBracket.bracket);
						tempLine.splice(0, 0, String.fromCharCode(0xFE76)); // prepending an arabic character ?!
						line = tempLine.join('');
					}
				}
				
			}
			
			// reverse back numeric sequence
			v = 0;
			while (v<numLineChars) {
				if (isNumeric(numbericString+line.charAt(v))) {
					numbericString += line.charAt(v);
				}
				v++;
			}
			if (numbericString.indexOf(numericSeparator) != -1) {
				tempLine = numbericString.split(numericSeparator);
				tempLine.reverse();
				line = line.split(numbericString).join(tempLine.join(numericSeparator));
			}
			
			if (string != '') {
				string += '\n';
			}
			string += line;
			i++;
		}
		
		return string;
		
	}
	
	private function getSingleBracket(input:String):Object {
		
		var numChars:Number = input.length;
		var code:Number;
		var numBrackets:Number = 0;
		var bracket:String;
		var isClosed:Boolean = false;
		var i:Number = 0;
		while (i<numChars) {
			code = input.charCodeAt(i);
			if (code == 40 || code == 91 || code == 123 || code == 171 || code == 10216 || 
				code == 41 || code == 93 || code == 125 || code == 187 || code == 10217) {
				bracket = input.charAt(i);
				if (code == 41 || code == 93 || code == 125 || code == 187 || code == 10217) {
					isClosed = true;
				}
				numBrackets++;
			}
			if (numBrackets>1) {
				bracket = null;
				break;
			}
			i++;
		}
		
		return {bracket:bracket, closed:isClosed};
		
	}
	
	private function getHTMLFormat(input:String, format:TextFormat):String {
		
		var htmlFormat:String = '<textformat leftmargin="'+((format.leftMargin) ? format.leftMargin : 0)+'" rightmargin="'+((format.rightMargin) ? format.rightMargin : 0)+'" leading="'+((format.leading) ? format.leading : 0)+'"><p align="'+((format.align) ? format.align : "left")+'"><font face="'+((format.font) ? format.font : "Arial")+'" color="'+((format.color) ? getProperColor(format.color) : "#000000")+'" size="'+((format.size) ? format.size : 12)+'" letterspacing="'+((format.letterSpacing) ? format.letterSpacing : 0)+'" kerning="'+((format.kerning) ? format.kerning : 0)+'">';
		
		htmlFormat += input;
		
		htmlFormat += "</font></p></textformat>";
		
		return htmlFormat;
		
	}
	
	private function getProperColor(input:Number):String {
		
		var color:String = '#';
		var checker:Number = 0;
		
		if (input.toString(16).length<6) {
			while (checker<6-input.toString(16).length) {
				color += '0';
				checker++;
			}
		}
		
		color += input.toString(16);
		
		return color;
		
	}
	
	private function isNumeric(num:String):Boolean {
		
		return !isNaN(parseInt(num));
		
	}
	
	private function convertNumbers(input:String):String {
		var string:String = "";
		var i:Number = 0;
		while (i<input.length) {
			switch (input.charAt(i)) {
				case '0':
				string += String.fromCharCode(0x0660);
				break;
				case '1':
				string += String.fromCharCode(0x0661);
				break;
				case '2':
				string += String.fromCharCode(0x0662);
				break;
				case '3':
				string += String.fromCharCode(0x0663);
				break;
				case '4':
				string += String.fromCharCode(0x0664);
				break;
				case '5':
				string += String.fromCharCode(0x0665);
				break;
				case '6':
				string += String.fromCharCode(0x0666);
				break;
				case '7':
				string += String.fromCharCode(0x0667);
				break;
				case '8':
				string += String.fromCharCode(0x0668);
				break;
				case '9':
				string += String.fromCharCode(0x0669);
				break;
				default:
				string += input.charAt(i);
				break;
			}
			i++;
		}
		return string;
	}
}